import re
from graphql.language.ast import Field, ObjectField, ObjectValue, StringValue, Name, Argument, FragmentSpread, InlineFragment, SelectionSet, Document, OperationDefinition, VariableDefinition
from graphql.language.visitor import Visitor
from graphql.language import ast
from graphql import parse, print_ast


def find_page_info(json_data):
    """Func to find the pageInfo from json response

    Args:
        json_data (dict): api response

    Returns:
        json: pageInfo
    """
    if isinstance(json_data, dict):
        if 'pageInfo' in json_data:
            return json_data['pageInfo']
        else:
            for value in json_data.values():
                page_info = find_page_info(value)
                if page_info:
                    return page_info
    elif isinstance(json_data, list):
        page_infos = []
        for item in json_data[0]:
            page_infos.append(find_page_info(json_data[0][item]))
        if page_infos:
            return page_infos
    return None


def modify_query_with_cursor(query, key, cursor):
    """Modify the GraphQL query by adding or replacing the cursor input for the specified key."""
    pattern = rf'(\b{key}\b[^}}]+cursor:\s")[^"]+'
    modified_query = re.sub(pattern, rf'\g<1>{cursor}', query)
    return modified_query


def create_object_field(name, value):
    """
    Func to create a object
    Args:
        name: name of the object
        value: value of the object

    Returns:

    """
    return ObjectField(name=Name(value=name), value=StringValue(value=value))


def add_cursor_to_input_field(ast, field_name, cursor_value):
    """
    Func to add cursor to input fields
    Args:
        ast: parsed ast
        field_name: field_name where to add
        cursor_value: cursor value

    Returns:

    """
    for definition in ast.definitions:
        if definition.operation == "query":
            for selection in definition.selection_set.selections:
                if isinstance(selection, Field):
                    if selection.name.value == field_name or (selection.alias and
                    selection.alias.value == field_name):
                        arguments = selection.arguments
                        if not arguments:
                            # If no arguments are present, create a new input argument
                            # with the cursor field
                            argument = Argument(
                                name=Name(value='input'),
                                value=ObjectValue(fields=[create_object_field('cursor',
                                cursor_value)])
                            )
                            selection.arguments.append(argument)
                        else:
                            # If input argument exists, check if it already has a cursor field
                            for argument in arguments:
                                if argument.name.value == 'input':
                                    input_value = argument.value
                                    for field in input_value.fields:
                                        if isinstance(field, ObjectField) and field.name.value == 'cursor':
                                            # If cursor field exists, replace its value
                                            field.value = StringValue(value=cursor_value)
                                            return  # Exit the loop if cursor field is found
                                    # If cursor field is not found, add it to the input fields
                                    cursor_field = create_object_field('cursor', cursor_value)
                                    input_value.fields.append(cursor_field)
                                    return  # Exit the loop if input argument is found
                        return


def replace_cursor_value(ast, key, cursor_value):
    """
    Func to replace cursor value
    Args:
        ast: parsed ast
        key: key to replace
        cursor_value: new cursor value

    Returns:

    """
    for definition in ast.definitions:
        if definition.operation == "query":
            for selection in definition.selection_set.selections:
                if isinstance(selection, Field) and selection.alias and selection.alias.value == key:
                    for argument in selection.arguments:
                        if argument.name.value == 'input':
                            input_value = argument.value
                            for field in input_value.fields:
                                if isinstance(field, ObjectField) and field.name.value == 'cursor':
                                    field.value = StringValue(value=cursor_value)
                                    return  # Exit the loop if cursor field is found
                            # If cursor field is not found, add it to the input fields
                            cursor_field = create_object_field('cursor', cursor_value)
                            input_value.fields.append(cursor_field)
                            return  # Exit the loop if input argument is found


def has_cursor(ast, key):
    """
    Func to check if dict has cursor
    Args:
        ast: parsed ast
        key: for which key to check

    Returns:

    """
    for definition in ast.definitions:
        if definition.operation == "query":
            for selection in definition.selection_set.selections:
                if isinstance(selection, Field) and selection.alias and selection.alias.value == key:
                    for argument in selection.arguments:
                        if argument.name.value == 'input':
                            input_value = argument.value
                            if isinstance(input_value, ObjectValue):
                                for field in input_value.fields:
                                    if field.name.value == 'cursor':
                                        return True
    return False

class RemoveQueryByStartingName(Visitor):
    """Class to remove queries from a multi-query that do not have next or prevCursor based on field names or aliases"""

    def __init__(self, query_start):
        self.query_start = query_start

    def enter_OperationDefinition(self, node, key, parent, path, ancestors):
        if node.operation == 'query':
            selections = node.selection_set.selections
            selections[:] = [selection for selection in selections if not self._should_remove_query(selection)]
            if not selections:
                # Remove variables when there are no remaining selections
                node.variable_definitions = None

    def _should_remove_query(self, selection):
        if isinstance(selection, FragmentSpread):
            return False
        if isinstance(selection, InlineFragment):
            return self._should_remove_query(selection.selection_set)
        if isinstance(selection, Field):
            field_name = selection.name.value
            alias_name = selection.alias.value if selection.alias else None
            if field_name.startswith(self.query_start) or (alias_name and alias_name.startswith(self.query_start)):
                return True
        return False

def add_page_info_to_queries(graphql_document):
    """Func to add page info to the graphql query

    Args:
        graphql_document (ast): parsed query

    Returns:
        str: page info added query
    """
    parsed_document = parse(graphql_document)
    modified_document = _add_page_info_to_queries(parsed_document)
    return print_ast(modified_document)

def _add_page_info_to_queries(node):
    if isinstance(node, Document):
        node.definitions = [_add_page_info_to_queries(definition) for definition in node.definitions]
    elif isinstance(node, Field):
        if node.selection_set is None:
            node.selection_set = SelectionSet(selections=[])
        node.selection_set.selections.append(Field(
            name=Name(value="pageInfo"),
            selection_set=SelectionSet(selections=[
                Field(name=Name(value="nextCursor")),
                Field(name=Name(value="prevCursor"))
            ])
        ))
    elif hasattr(node, "selection_set"):
        node.selection_set.selections = [_add_page_info_to_queries(selection) for selection in node.selection_set.selections]
    return node

def remove_unused_variables(document_ast, query):
    """Func to remove unused variables from the query

    Args:
        document_ast (ast): parsed query
        query (str): query
    """
    for definition in document_ast.definitions:
        for _count, _variable in enumerate(definition.variable_definitions):
            if query.count(_variable.variable.name.value) == 1:
                del document_ast.definitions[0].variable_definitions[_count]
    return print_ast(document_ast)



"""
Module: constant.py
Description: This module contains the constant used in airstack sdk.
"""

class AirstackConstants:
    """Class for keeping constants
    """
    API_ENDPOINT_PROD = 'https://api.airstack.xyz/gql'
    API_TIMEOUT = 60
    SUCCESS_STATUS_CODE = 200
    UNPROCESSABLE_STATUS_CODE = 422



"""
Module: execute_query.py
Description: This module contains the methods to execute the query.
"""

import json
import re
from graphql import parse, print_ast, visit
from airstack.send_request import SendRequest
from graphql.language.ast import OperationDefinition
from airstack.generic import (
    find_page_info,
    add_cursor_to_input_field,
    replace_cursor_value,
    has_cursor,
    RemoveQueryByStartingName,
    add_page_info_to_queries,
    remove_unused_variables
)
from airstack.constant import AirstackConstants


class QueryResponse:
    """Class for generate the query response
    """
    def __init__(self, response, status_code, error, has_next_page=None, has_prev_page=None,
    get_next_page=None, get_prev_page=None):
        """Init function

        Args:
            response (dic): query response
            status_code (int): status code
            error (str): error if there
            has_next_page (bool, optional): if next page is there. Defaults to None.
            has_prev_page (bool, optional): if previous page is there. Defaults to None.
            get_next_page (func, optional): func to get the next page data. Defaults to None.
            get_prev_page (func, optional): func to get the previous page data. Defaults to None.
        """
        self.data = response
        self.status_code = status_code
        self.error = error
        self.has_next_page = has_next_page
        self.has_prev_page = has_prev_page
        self.get_next_page = get_next_page
        self.get_prev_page = get_prev_page

class AirstackClient:
    """Class to create api client for airstack api's
    """

    def __init__(self, url=None, api_key=None):
        """Init function for api client

        Args:
            url (str, optional): base url for server. Defaults to None.
            api_key (str, required): api key. Defaults to None.

        Raises:
            ValueError: _description_
        """
        self.url = AirstackConstants.API_ENDPOINT_PROD if url is None else url
        if api_key is None:
            raise ValueError("API key must be provided.")

        self.timeout = AirstackConstants.API_TIMEOUT
        self.api_key = api_key
    
    def create_execute_query_object(self, query=None, variables=None):
        # Create a new object using the caller's `self` variables
        execute_query = ExecuteQuery(query=query, variables=variables, url=self.url, api_key=self.api_key, timeout=self.timeout)
        return execute_query

class ExecuteQuery:
    """_summary_

    Returns:
        _type_: _description_
    """
    def __init__(self, query=None, variables=None, url=None, api_key=None, timeout=None):
        self.deleted_queries = []
        self.query = query
        self.variables = variables
        self.url = url
        self.api_key = api_key
        self.timeout = timeout

    async def execute_query(self, query=None):
        """Async function to run a GraphQL query and get the data

        Args:
            query (str): GraphQL query string. Defaults to None
            variables (dict, optional): Variables for the query. Defaults to
            None.

        Returns:
            Tuple: GraphQL response data or None, GraphQL response status code,
            error message or None
        """
        if query is None:
            query = self.query
            
        headers = {
            'Content-Type': 'application/json',
            'Authorization': self.api_key
        }
        payload = {
            'query': query,
            'variables': self.variables
        }

        response, status_code, error = await SendRequest.send_post_request(
            url=self.url, headers=headers, data=json.dumps(payload), timeout=self.timeout)
        return QueryResponse(response, status_code, error)

    async def execute_paginated_query(self, query=None, variables=None):
        """Async function to execute paginated query.

        Args:
            query (str): GraphQL query string. Defaults to None
            variables (dict, optional): Variables for the query. Defaults to
            None.

        Returns:
            Tuple: GraphQL response data or None, GraphQL response status code,
            error message or None, next cursor,
            previous cursor
        """
        if query is None:
            query = self.query

        regex = re.compile(r'pageInfo')
        has_page_info = regex.search(query)
        if has_page_info is None:
            query = add_page_info_to_queries(query)

        query_response = await self.execute_query(query=query)
        if query_response.error is not None:
            return QueryResponse(None, query_response.status_code, query_response.error, None, None, None, None)

        page_info = {}
        for _key, value in query_response.data.items():
            page_info[_key] = find_page_info(query_response.data[_key])

        has_next_page = any(page_info['nextCursor'] != '' for page_info in page_info.values())
        has_prev_page = any(page_info['prevCursor'] != '' for page_info in page_info.values())

        return QueryResponse(query_response.data, query_response.status_code,
                            query_response.error, has_next_page, has_prev_page,
                            self.get_next_page(query, variables, page_info),
                            self.get_prev_page(query, variables, page_info))

    async def get_next_page(self, query, variables, page_info):
        """Async function to get the next page data.

        Args:
            query (str): GraphQL query string.
            variables (dict): Variables for the query.
            page_info (dict): Page info dictionary.
            deleted_queries (list): List to store deleted queries.
            next_cursors (list): List to store next cursors.

        Returns:
            Tuple: GraphQL response data or None, GraphQL response status code,
            error message or None, next cursor,
            previous cursor
        """
        next_query = query
        stored = False
        for _page_info_key, _page_info_value in page_info.items():
            document_ast = parse(next_query)
            if _page_info_value['nextCursor'] == "":
                self.deleted_queries.append(next_query)
                stored = True
                visitor = RemoveQueryByStartingName(query_start=_page_info_key)
                document_ast = visit(document_ast, visitor)
                next_query = remove_unused_variables(document_ast=document_ast, query=print_ast(document_ast))
            else:
                if not stored:
                    self.deleted_queries.append(None)
                if has_cursor(document_ast, _page_info_key):
                    replace_cursor_value(document_ast, _page_info_key, _page_info_value['nextCursor'])
                else:
                    add_cursor_to_input_field(document_ast, _page_info_key, _page_info_value['nextCursor'])
                next_query = print_ast(document_ast)
        return await self.execute_paginated_query(next_query, variables)

    async def get_prev_page(self, query, variables, page_info):
        """Async function to get the previous page data.

        Args:
            query (str): GraphQL query string.
            variables (dict): Variables for the query.
            page_info (dict): Page info dictionary.

            Returns:
                Tuple: GraphQL response data or None, GraphQL response status code,
                error message or None, next cursor,
                previous cursor
            """
        deleted_query = self.deleted_queries.pop()
        if deleted_query:
            next_query = deleted_query
        else:
            next_query = query
        for _page_info_key, _page_info_value in page_info.items():
            document_ast = parse(next_query)
            if has_cursor(document_ast, _page_info_key):
                replace_cursor_value(document_ast, _page_info_key,
                _page_info_value['prevCursor'])
            else:
                add_cursor_to_input_field(document_ast, _page_info_key,
                _page_info_value['prevCursor'])
            next_query = print_ast(document_ast)
        return await self.execute_paginated_query(next_query, variables)



__author__ = 'sarvesh.singh'

import json
import aiohttp
from airstack.constant import AirstackConstants


class SendRequest:
    """
    Send Request
    """

    @staticmethod
    async def send_post_request(url=None, headers=None, data=None,
                                timeout=True):
        """Async function to send post request

        Args:
            url (str, optional): server url. Defaults to None.
            headers (dict, optional): headers. Defaults to None.
            data (dict, optional): json request body. Defaults to None.
            timeout (int, optional): timeout for api. Defaults to True.

        Returns:
            Tuple: JSON response or None, response status code, error message or None
        """
        async with aiohttp.ClientSession() as session:
            try:
                async with session.post(url=url, headers=headers,
                                        data=data,
                                        timeout=timeout) as response:
                    content = await response.text()
                    try:
                        nt = json.loads(content)
                    except json.JSONDecodeError:
                        nt = json.loads(json.dumps(content))

                    if response.status != AirstackConstants.SUCCESS_STATUS_CODE:
                        if response.status == AirstackConstants.UNPROCESSABLE_STATUS_CODE:
                            return None, response.status, response.reason
                        return None, response.status, nt["error"]

                    if "errors" in nt:
                        return nt, response.status, nt["errors"]

                    return nt["data"], response.status, None
            except Exception as exec:
                return None, response.status, str(exec)
